# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

import os
import random
from typing import Tuple

import pytest
import torch

from xformers.ops import (
    sequence_parallel_leading_matmul,
    sequence_parallel_trailing_matmul,
)

from .multiprocessing_utils import launch_subprocesses

compute_capability = (0, 0)
if torch.cuda.is_available():
    compute_capability = torch.cuda.get_device_capability("cuda")
cuda_sm80_only = pytest.mark.skipif(
    compute_capability < (8, 0), reason="requires sm70+"
)
at_least_2_gpus = pytest.mark.skipif(
    torch.cuda.device_count() < 2, reason="needs at least 2 GPUs"
)


def reference_leading(input_, w1, w2):
    hidden1 = torch.matmul(input_, w1.t())
    hidden2 = torch.matmul(input_, w2.t())
    return [hidden1, hidden2]


def reference_trailing(hidden, w):
    output = torch.matmul(hidden, w.t())
    return output


def xformers_leading(input_, w1, w2, *, fuse, group):
    return sequence_parallel_leading_matmul(
        input_, [w1.t(), w2.t()], fuse=fuse, process_group=group
    )


def xformers_trailing(hidden, w, *, fuse, group):
    return sequence_parallel_trailing_matmul(
        hidden, w.t(), fuse=fuse, process_group=group
    )


def inner_seqpar(
    kind: str,
    step: str,
    dims: Tuple[int, ...],
    dtype: torch.dtype,
    compile: bool,
    seed: int,
):
    os.environ["TORCH_SYMM_MEM_ALLOW_OVERLAPPING_DEVICES"] = "1"

    my_rank = torch.distributed.get_rank()
    world_size = torch.distributed.get_world_size()
    subgroup = torch.distributed.new_group()

    fused = True
    if kind == "unfused":
        fused = False
    elif kind == "fallback":
        os.environ["DISABLE_FUSED_SEQUENCE_PARALLEL"] = "1"

    torch.random.manual_seed(seed)
    torch._dynamo.reset_code_caches()  # avoids hitting recompilation limit

    batch_dims = dims[:-2]
    outer_dim = dims[-2]
    inner_dim = dims[-1]

    # To check for correctness we want to compare the outputs but the accuracy
    # of matmuls, apparently, is not that great. We thus try to produce inputs
    # for which no rounding at all will occur. We do this by using zero or one
    # inputs, so their product will also be zero or one, and keep the reduction
    # dimension small enough so that they fit in the mantissa without overflow.
    max_exact_value = 2 * (1 / torch.finfo(dtype).eps)
    # 0.25 is the ratio of expected ones and we aim at 2/3 of the safe range
    assert outer_dim * 0.25 <= max_exact_value * 0.66
    assert inner_dim * world_size * 0.25 <= max_exact_value * 0.66

    def my_chunk(t, *, dim):
        return t.tensor_split(world_size, dim=dim)[my_rank]

    if step == "leading":
        input_ = torch.testing.make_tensor(
            batch_dims + (outer_dim,),
            dtype=dtype,
            device="cuda",
            low=0,
            high=1,
        ).round()
        weight1, weight2 = [
            torch.testing.make_tensor(
                (inner_dim * (idx + 1), outer_dim),
                dtype=dtype,
                device="cuda",
                low=0,
                high=1,
            ).round()
            for idx in range(2)
        ]
        gradient1, gradient2 = [
            torch.testing.make_tensor(
                batch_dims + (inner_dim * (idx + 1),),
                dtype=dtype,
                device="cuda",
                low=0,
                high=1,
            ).round()
            for idx in range(2)
        ]

        # Non-fused reference code
        input_ref = input_.detach().requires_grad_()
        weight1_ref = weight1.detach().requires_grad_()
        weight2_ref = weight2.detach().requires_grad_()

        output1_ref, output2_ref = reference_leading(
            input_ref, weight1_ref, weight2_ref
        )
        torch.autograd.backward([output1_ref, output2_ref], [gradient1, gradient2])

        my_output1_ref = my_chunk(output1_ref, dim=-1)
        my_output2_ref = my_chunk(output2_ref, dim=-1)
        my_weight1_grad_ref = my_chunk(weight1_ref.grad, dim=0)
        my_weight2_grad_ref = my_chunk(weight2_ref.grad, dim=0)
        my_input_grad_ref = my_chunk(input_ref.grad, dim=0)

        # Faster fused mode
        my_input_xf = my_chunk(input_, dim=0).detach().requires_grad_()
        my_weight1_xf = my_chunk(weight1, dim=0).detach().requires_grad_()
        my_weight2_xf = my_chunk(weight2, dim=0).detach().requires_grad_()
        my_gradient1 = my_chunk(gradient1, dim=-1)
        my_gradient2 = my_chunk(gradient2, dim=-1)

        my_output1_xf, my_output2_xf = torch.compile(
            xformers_leading, fullgraph=True, disable=not compile
        )(my_input_xf, my_weight1_xf, my_weight2_xf, fuse=fused, group=subgroup)
        torch.autograd.backward(
            [my_output1_xf, my_output2_xf], [my_gradient1, my_gradient2]
        )

        my_weight1_grad_xf = my_weight1_xf.grad
        my_weight2_grad_xf = my_weight2_xf.grad
        my_input_grad_xf = my_input_xf.grad

        # Checks
        torch.testing.assert_close(my_output1_ref, my_output1_xf)
        torch.testing.assert_close(my_output2_ref, my_output2_xf)
        torch.testing.assert_close(my_input_grad_ref, my_input_grad_xf)
        torch.testing.assert_close(my_weight1_grad_ref, my_weight1_grad_xf)
        torch.testing.assert_close(my_weight2_grad_ref, my_weight2_grad_xf)

        torch.library.opcheck(
            torch.ops.xformers_python.sequence_parallel_leading_matmul_fwd,
            (my_input_xf.flatten(0, -2), [my_weight1_xf.t(), my_weight2_xf.t()]),
            {"fuse": fused, "process_group_name": subgroup.group_name},
        )

    elif step == "trailing":
        input_ = torch.testing.make_tensor(
            batch_dims + (inner_dim,),
            dtype=dtype,
            device="cuda",
            low=0,
            high=1,
        ).round()
        weight = torch.testing.make_tensor(
            (outer_dim, inner_dim),
            dtype=dtype,
            device="cuda",
            low=0,
            high=1,
        ).round()
        gradient = torch.testing.make_tensor(
            batch_dims + (outer_dim,),
            dtype=dtype,
            device="cuda",
            low=0,
            high=1,
        ).round()

        # Non-fused reference code
        input_ref = input_.detach().requires_grad_()
        weight_ref = weight.detach().requires_grad_()

        output_ref = reference_trailing(input_ref, weight_ref)
        torch.autograd.backward([output_ref], [gradient])

        my_output_ref = my_chunk(output_ref, dim=0)
        my_weight_grad_ref = my_chunk(weight_ref.grad, dim=1)
        my_input_grad_ref = my_chunk(input_ref.grad, dim=-1)

        # Faster fused mode
        my_input_xf = my_chunk(input_, dim=-1).detach().clone().requires_grad_()
        my_weight_xf = my_chunk(weight, dim=1).detach().requires_grad_()
        my_gradient = my_chunk(gradient, dim=0)

        my_output_xf = torch.compile(
            xformers_trailing, fullgraph=True, disable=not compile
        )(my_input_xf, my_weight_xf, fuse=fused, group=subgroup)
        torch.autograd.backward([my_output_xf], [my_gradient])

        my_weight_grad_xf = my_weight_xf.grad
        my_input_grad_xf = my_input_xf.grad

        # Checks
        torch.testing.assert_close(my_output_ref, my_output_xf)
        torch.testing.assert_close(my_input_grad_ref, my_input_grad_xf)
        torch.testing.assert_close(my_weight_grad_ref, my_weight_grad_xf)

        torch.library.opcheck(
            torch.ops.xformers_python.sequence_parallel_trailing_matmul_fwd,
            (my_input_xf.flatten(0, -2), my_weight_xf.t()),
            {"fuse": fused, "process_group_name": subgroup.group_name},
        )


# PyTorch doesn't support pre-sm80 for its signaling kernels
# https://github.com/pytorch/pytorch/pull/146308
@cuda_sm80_only
@pytest.mark.parametrize(
    "kind",
    [
        "singleton",
        pytest.param("unfused", marks=at_least_2_gpus),
        pytest.param("fallback", marks=at_least_2_gpus),
        "fused",
    ],
)
@pytest.mark.parametrize(
    "step",
    [
        "leading",
        "trailing",
    ],
)
@pytest.mark.parametrize(
    "dims",
    [
        pytest.param((2, 2, 512, 512, 256), id="nice-shapes"),
        pytest.param((2, 1023, 511, 257), id="ugly-shapes"),
    ],
)
@pytest.mark.parametrize(
    "dtype",
    [
        pytest.param(torch.bfloat16, id="bf16"),
        pytest.param(torch.float16, id="fp16"),
        pytest.param(torch.float32, id="fp32"),
    ],
)
@pytest.mark.parametrize(
    "compile", [pytest.param(False, id="eager"), pytest.param(True, id="compile")]
)
def test_seqpar(
    kind: str,
    step: str,
    dims: Tuple[int, ...],
    dtype: torch.dtype,
    compile: bool,
):
    if compile and dtype is torch.bfloat16 and compute_capability < (8, 0):
        # https://fb.workplace.com/groups/1075192433118967/posts/1480158559289017
        pytest.skip("Dynamo misbehaves on V100 or earlier when handling bf16")

    world_size = 1 if kind == "singleton" else 2
    seed = random.getrandbits(32)
    launch_subprocesses(
        world_size=world_size,
        fn=inner_seqpar,
        kind=kind,
        step=step,
        dims=dims,
        dtype=dtype,
        compile=compile,
        seed=seed,
    )



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from typing import Callable, List, Tuple

import torch
from torch.distributed.distributed_c10d import _resolve_process_group, GroupName

from .differentiable_collectives import (
    gather_along_first_dim,
    gather_along_first_dim_async,
    reduce_scatter_along_first_dim,
    reduce_scatter_along_first_dim_async,
)
from .sequence_parallel_fused_ops import (
    fused_allgather_and_anything,
    fused_allgather_and_linear,
    fused_anything_and_reducescatter,
    fused_linear_and_reducescatter,
)
from .tiled_matmul import tiled_matmul, tiled_matmul_out


@torch.library.custom_op(
    "xformers_python::sequence_parallel_leading_matmul_fwd",
    mutates_args=(),
    device_types="cuda",
)
def sequence_parallel_leading_matmul_fwd(
    scattered_input: torch.Tensor,
    weights: List[torch.Tensor],
    fuse: bool,
    process_group_name: GroupName,
) -> List[torch.Tensor]:
    process_group = _resolve_process_group(process_group_name)

    if fuse:
        gathered_outputs = fused_allgather_and_linear(
            scattered_input, [w.t() for w in weights], group=process_group
        )
    else:
        gathered_input = gather_along_first_dim(
            scattered_input, process_group=process_group
        )
        (gathered_outputs,) = tiled_matmul(
            [[gathered_input]],
            [[w for w in weights]],
        )
    return gathered_outputs


@torch.library.register_fake("xformers_python::sequence_parallel_leading_matmul_fwd")
def sequence_parallel_leading_matmul_fwd_fake(
    scattered_input: torch.Tensor,
    weights: List[torch.Tensor],
    fuse: bool,
    process_group_name: GroupName,
) -> List[torch.Tensor]:
    mp_size = _resolve_process_group(process_group_name).size()
    return [
        scattered_input.new_empty((scattered_input.shape[0] * mp_size, w.shape[1]))
        for w in weights
    ]


@torch.library.custom_op(
    "xformers_python::sequence_parallel_leading_matmul_bwd",
    mutates_args=(),
    device_types="cuda",
)
def sequence_parallel_leading_matmul_bwd(
    scattered_input: torch.Tensor,
    weights: List[torch.Tensor],
    grad_gathered_outputs: List[torch.Tensor],
    fuse: bool,
    process_group_name: GroupName,
) -> Tuple[torch.Tensor, List[torch.Tensor]]:
    process_group = _resolve_process_group(process_group_name)
    mp_size = process_group.size()

    # torch.library.opcheck gives us gradients whose strides are zero.
    # See https://github.com/pytorch/pytorch/issues/132857.
    grad_gathered_outputs = [
        grad_go.clone() if any(s == 0 for s in grad_go.stride()) else grad_go
        for grad_go in grad_gathered_outputs
    ]

    if fuse:
        grad_scattered_input = torch.empty_like(scattered_input)
        grad_weights = [torch.zeros_like(w) for w in weights]

        grad_gathered_outputss = [
            grad_go.tensor_split(mp_size, dim=0) for grad_go in grad_gathered_outputs
        ]

        def my_si_matmul(
            grad_gathered_inputs: List[torch.Tensor],
            dst_rank: int,
            stream_factory: Callable[[], torch.cuda.Stream],
        ) -> None:
            (grad_gi,) = grad_gathered_inputs
            with torch.cuda.stream(stream_factory()):
                tiled_matmul_out(
                    [[grad_gos[dst_rank] for grad_gos in grad_gathered_outputss]],
                    [[w.t()] for w in weights],
                    out=[[grad_gi]],
                )

        fused_anything_and_reducescatter(
            my_si_matmul,
            [grad_scattered_input],
            group=process_group,
        )

        # Each pair of shards of input and grad_output accumulates into the same
        # grad_weight. Thus we need to make sure that the in-place addmms are
        # sequenced correctly for each of the grad_weights.
        events = [torch.cuda.Event() for _ in weights]

        def my_w_matmul(
            gathered_inputs_shard: List[torch.Tensor],
            src_rank: int,
            stream_factory: Callable[[], torch.cuda.Stream],
        ) -> None:
            (gi_shard,) = gathered_inputs_shard
            for grad_gos, grad_w, event in zip(
                grad_gathered_outputss, grad_weights, events
            ):
                with torch.cuda.stream(stream_factory()):
                    event.wait()
                    grad_w.t().addmm_(grad_gos[src_rank].t(), gi_shard)
                    event.record()

        fused_allgather_and_anything(
            [scattered_input],
            my_w_matmul,
            group=process_group,
        )
    else:
        gathered_input, handle = gather_along_first_dim_async(
            scattered_input, process_group=process_group
        )
        ((grad_gathered_input,),) = tiled_matmul(
            [[grad_go for grad_go in grad_gathered_outputs]],
            [[w.t()] for w in weights],
        )
        if handle is not None:
            handle.wait()

        grad_scattered_input, handle = reduce_scatter_along_first_dim_async(
            grad_gathered_input, process_group=process_group
        )

        grad_weights_tuples = tiled_matmul(
            [[grad_go.t()] for grad_go in grad_gathered_outputs],
            [[gathered_input]],
        )
        if handle is not None:
            handle.wait()

        grad_weights = [grad_w.t() for (grad_w,) in grad_weights_tuples]

    return grad_scattered_input, grad_weights


@torch.library.register_fake("xformers_python::sequence_parallel_leading_matmul_bwd")
def sequence_parallel_leading_matmul_bwd_fake(
    scattered_input: torch.Tensor,
    weights: List[torch.Tensor],
    grad_gathered_outputs: List[torch.Tensor],
    fuse: bool,
    process_group_name: GroupName,
) -> Tuple[torch.Tensor, List[torch.Tensor]]:
    return (torch.empty_like(scattered_input), [torch.empty_like(w) for w in weights])


def sequence_parallel_leading_matmul_setup_context(ctx, inputs, output):
    scattered_input, weights, fuse, process_group_name = inputs
    ctx.save_for_backward(scattered_input, *weights)
    ctx.fuse = fuse
    ctx.process_group_name = process_group_name


def sequence_parallel_leading_matmul_bwd_bridge(ctx, grad_gathered_outputs):
    scattered_input, *weights = ctx.saved_tensors
    (
        grad_scattered_input,
        grad_weights,
    ) = sequence_parallel_leading_matmul_bwd(
        scattered_input,
        list(weights),
        list(grad_gathered_outputs),
        ctx.fuse,
        ctx.process_group_name,
    )
    return grad_scattered_input, grad_weights, None, None


torch.library.register_autograd(
    "xformers_python::sequence_parallel_leading_matmul_fwd",
    sequence_parallel_leading_matmul_bwd_bridge,
    setup_context=sequence_parallel_leading_matmul_setup_context,
)


def sequence_parallel_leading_matmul(
    x: torch.Tensor,
    ws: List[torch.Tensor],
    *,
    fuse: bool,
    process_group: torch.distributed.ProcessGroup,
) -> List[torch.Tensor]:
    os = sequence_parallel_leading_matmul_fwd(
        x.flatten(0, -2), ws, fuse, process_group.group_name
    )
    return [o.view(-1, *x.shape[1:-1], w.shape[1]) for o, w in zip(os, ws)]


@torch.library.custom_op(
    "xformers_python::sequence_parallel_trailing_matmul_fwd",
    mutates_args=(),
    device_types="cuda",
)
def sequence_parallel_trailing_matmul_fwd(
    gathered_input: torch.Tensor,
    weight: torch.Tensor,
    fuse: bool,
    process_group_name: GroupName,
) -> torch.Tensor:
    process_group = _resolve_process_group(process_group_name)

    if fuse:
        scattered_output = fused_linear_and_reducescatter(
            gathered_input, weight.t(), group=process_group
        )
    else:
        gathered_output = torch.matmul(gathered_input, weight)
        scattered_output = reduce_scatter_along_first_dim(
            gathered_output, process_group=process_group
        )
    return scattered_output


@torch.library.register_fake("xformers_python::sequence_parallel_trailing_matmul_fwd")
def sequence_parallel_trailing_matmul_fwd_fake(
    gathered_input: torch.Tensor,
    weight: torch.Tensor,
    fuse: bool,
    process_group_name: GroupName,
) -> torch.Tensor:
    mp_size = _resolve_process_group(process_group_name).size()
    return gathered_input.new_empty(
        (gathered_input.shape[0] // mp_size, weight.shape[1])
    )


@torch.library.custom_op(
    "xformers_python::sequence_parallel_trailing_matmul_bwd",
    mutates_args=(),
    device_types="cuda",
)
def sequence_parallel_trailing_matmul_bwd(
    gathered_input: torch.Tensor,
    weight: torch.Tensor,
    grad_scattered_output: torch.Tensor,
    fuse: bool,
    process_group_name: GroupName,
) -> Tuple[torch.Tensor, torch.Tensor]:
    process_group = _resolve_process_group(process_group_name)
    mp_size = process_group.size()

    # torch.library.opcheck gives us gradients whose strides are zero.
    # See https://github.com/pytorch/pytorch/issues/132857.
    if any(s == 0 for s in grad_scattered_output.stride()):
        grad_scattered_output = grad_scattered_output.clone()

    if fuse:
        grad_gathered_input = torch.empty_like(gathered_input)
        grad_weight = torch.zeros_like(weight)

        gathered_inputs = gathered_input.tensor_split(mp_size, dim=0)
        grad_gathered_inputs = grad_gathered_input.tensor_split(mp_size, dim=0)

        def my_gi_and_w_matmul(
            grad_gathered_outputs_shard: List[torch.Tensor],
            src_rank: int,
            stream_factory: Callable[[], torch.cuda.Stream],
        ) -> None:
            (grad_go_shard,) = grad_gathered_outputs_shard
            with torch.cuda.stream(stream_factory()):
                torch.matmul(
                    grad_go_shard, weight.t(), out=grad_gathered_inputs[src_rank]
                )
            with torch.cuda.stream(stream_factory()):
                grad_weight.t().addmm_(grad_go_shard.t(), gathered_inputs[src_rank])

        fused_allgather_and_anything(
            [grad_scattered_output],
            my_gi_and_w_matmul,
            group=process_group,
        )
    else:
        grad_gathered_output = gather_along_first_dim(
            grad_scattered_output, process_group=process_group
        )
        grad_gathered_input = torch.matmul(grad_gathered_output, weight.t())
        grad_weight = torch.matmul(grad_gathered_output.t(), gathered_input).t()

    return grad_gathered_input, grad_weight


@torch.library.register_fake("xformers_python::sequence_parallel_trailing_matmul_bwd")
def sequence_parallel_trailing_matmul_bwd_fake(
    gathered_input: torch.Tensor,
    weight: torch.Tensor,
    grad_scattered_output: torch.Tensor,
    fuse: bool,
    process_group_name: GroupName,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return (torch.empty_like(gathered_input), torch.empty_like(weight))


def sequence_parallel_trailing_matmul_setup_context(ctx, inputs, output):
    gathered_input, weight, fuse, process_group_name = inputs
    ctx.save_for_backward(gathered_input, weight)
    ctx.fuse = fuse
    ctx.process_group_name = process_group_name


def sequence_parallel_trailing_matmul_bwd_bridge(ctx, grad_scattered_output):
    gathered_input, weight = ctx.saved_tensors
    (
        grad_gathered_input,
        grad_weight,
    ) = sequence_parallel_trailing_matmul_bwd(
        gathered_input,
        weight,
        grad_scattered_output,
        ctx.fuse,
        ctx.process_group_name,
    )
    return grad_gathered_input, grad_weight, None, None


torch.library.register_autograd(
    "xformers_python::sequence_parallel_trailing_matmul_fwd",
    sequence_parallel_trailing_matmul_bwd_bridge,
    setup_context=sequence_parallel_trailing_matmul_setup_context,
)


def sequence_parallel_trailing_matmul(
    x: torch.Tensor,
    w: torch.Tensor,
    *,
    fuse: bool,
    process_group: torch.distributed.ProcessGroup,
) -> torch.Tensor:
    o = sequence_parallel_trailing_matmul_fwd(
        x.flatten(0, -2), w, fuse, process_group.group_name
    )
    return o.view(-1, *x.shape[1:-1], w.shape[1])



# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from typing import List, Optional, Set, Tuple, Union

import dcgm_fields
import torch
from dcgm_fields import DcgmFieldGetById
from dcgm_structs import DCGM_GROUP_EMPTY, DCGM_OPERATION_MODE_AUTO
from pydcgm import DcgmFieldGroup, DcgmGroup, DcgmHandle

from .profiler import _Profiler, logger


class DCGMProfiler:
    """Profiler that triggers start of DCGM profiler."""

    def __init__(
        self,
        main_profiler: "_Profiler",
        gpus_to_profile: Optional[Tuple[int, ...]] = None,
        field_ids_to_profile=(
            dcgm_fields.DCGM_FI_PROF_SM_ACTIVE,
            dcgm_fields.DCGM_FI_PROF_SM_OCCUPANCY,
            dcgm_fields.DCGM_FI_PROF_PIPE_TENSOR_ACTIVE,
            dcgm_fields.DCGM_FI_PROF_DRAM_ACTIVE,
            dcgm_fields.DCGM_FI_PROF_PCIE_TX_BYTES,
            dcgm_fields.DCGM_FI_PROF_PCIE_RX_BYTES,
            dcgm_fields.DCGM_FI_PROF_NVLINK_TX_BYTES,
            dcgm_fields.DCGM_FI_PROF_NVLINK_RX_BYTES,
        ),
        updateFreq: int = 5000,  # in microseconds
    ) -> None:
        """
        Args:
            main_profiler: The main profiler object.
            gpus_to_profile: A tuple of integers representing the GPUs to profile. If `None`,
                then the "default" GPU is used.
            field_ids_to_profile:
                See https://github.com/NVIDIA/DCGM/blob/master/testing/python3/dcgm_fields.py#L436
                for a full list of available fields. Note that not all fields are profilable.
            updateFreq: The interval of two consecutive updates of each field. Defaults to 5000 microseconds.
                This is a good tradeoff between performance and accuracy.
                An even smaller updateFreq is not supported well by A100.
                If the step to profile takes more than 5000 microseconds, then a larger updateFreq could also be used.
        """
        self.main_profiler = main_profiler
        self.updateFreq = updateFreq

        self.dcgmHandle = DcgmHandle(
            ipAddress="127.0.0.1", opMode=DCGM_OPERATION_MODE_AUTO
        )

        if gpus_to_profile is None:
            default_gpu: int = torch.empty([], device="cuda").device.index
            self.dcgmGroup = self.create_dcgm_group((default_gpu,))
        else:
            self.dcgmGroup = self.create_dcgm_group(gpus_to_profile)

        self.dcgmFieldGroup = self.create_profiling_field_group(field_ids_to_profile)

    def create_dcgm_group(
        self, gpus_to_profile: Union[Tuple[int], Tuple[int, ...]]
    ) -> Optional[DcgmGroup]:
        if self.dcgmHandle is None:
            return None

        dcgmSystem = self.dcgmHandle.GetSystem()
        supportedGPUs = dcgmSystem.discovery.GetAllSupportedGpuIds()

        valid_gpus_to_profile: List[int] = [
            gpu for gpu in gpus_to_profile if gpu in supportedGPUs
        ]
        if len(valid_gpus_to_profile) < 1:
            logger.warning(
                f"The provided GPUs are not supported on this system: "
                f"provided {gpus_to_profile}, supported {supportedGPUs}. "
                f"No data will be captured."
            )
            return None

        dcgmGroup = DcgmGroup(
            self.dcgmHandle,
            groupName="DCGMProfiler",
            groupType=DCGM_GROUP_EMPTY,
        )

        for gpu in valid_gpus_to_profile:
            dcgmGroup.AddGpu(gpu)

        return dcgmGroup

    def get_profilable_fields(self) -> Set[int]:
        assert self.dcgmGroup is not None

        dcgmMetricGroups = self.dcgmGroup.profiling.GetSupportedMetricGroups()
        profilableFieldIds = set()
        for group_idx in range(dcgmMetricGroups.numMetricGroups):
            metric_group = dcgmMetricGroups.metricGroups[group_idx]
            for field_id in metric_group.fieldIds[: metric_group.numFieldIds]:
                profilableFieldIds.add(field_id)
        return profilableFieldIds

    def create_profiling_field_group(
        self,
        fieldIdsToProfile: Optional[Tuple[int, ...]],
    ) -> Optional[DcgmFieldGroup]:
        if self.dcgmGroup is None:
            return None

        # Get all field ids that can be profiled.
        profilableFieldIds = self.get_profilable_fields()

        # Check which of the provided field ids are valid and invalid.
        if fieldIdsToProfile is None:
            validFieldIds = list(profilableFieldIds)
            invalidFieldIds = []
        else:
            validFieldIds = [
                field_id
                for field_id in fieldIdsToProfile
                if field_id in profilableFieldIds
            ]
            invalidFieldIds = [
                field_id
                for field_id in fieldIdsToProfile
                if field_id not in profilableFieldIds
            ]

        if not validFieldIds:
            logger.warning(
                "None of the provided field ids could be profiled.\n"
                f"  Provided: {fieldIdsToProfile}\n"
                f"  Supported: {profilableFieldIds}\n"
                "No data will be captured."
            )
            return None

        if invalidFieldIds:
            logger.warning(
                f"The following field ids cannot be profiled: {invalidFieldIds}. "
                f"Profiling {validFieldIds} only."
            )
        dcgmFieldGroup = DcgmFieldGroup(
            self.dcgmHandle, name="Profiling", fieldIds=validFieldIds
        )
        return dcgmFieldGroup

    def __enter__(self) -> None:
        if self.dcgmGroup is not None and self.dcgmFieldGroup is not None:
            self.dcgmGroup.samples.WatchFields(
                self.dcgmFieldGroup, self.updateFreq, 3600, 0
            )

            # Start collecting the profiling results run in background.
            self.profiling_results = self.dcgmGroup.samples.GetAllSinceLastCall(
                None, self.dcgmFieldGroup
            )

            # It is necessary to call GetAllSinceLastCall and EmptyValues twice
            # to clear old data from previous profilings
            # (otherwise the new profiling data is appended to the old data from previous profiling).
            self.dcgmGroup.samples.GetAllSinceLastCall(
                self.profiling_results, self.dcgmFieldGroup
            )
            self.profiling_results.EmptyValues()

            self.dcgmGroup.samples.GetAllSinceLastCall(
                self.profiling_results, self.dcgmFieldGroup
            )
            self.profiling_results.EmptyValues()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.dcgmGroup is not None and self.dcgmFieldGroup is not None:
            self.dcgmGroup.samples.UnwatchFields(self.dcgmFieldGroup)

            # Delete the group.
            self.dcgmGroup.Delete()
            del self.dcgmGroup
            self.dcgmGroup = None

            # Disconnect from the hostengine by deleting the DcgmHandle object.
            del self.dcgmHandle
            self.dcgmHandle = None

    def step(self) -> None:
        if self.dcgmGroup is not None and self.dcgmFieldGroup is not None:
            # Collect the profiling results.
            self.dcgmGroup.samples.GetAllSinceLastCall(
                self.profiling_results, self.dcgmFieldGroup
            )

            # Save profiling results to log.
            for gpu_id in self.profiling_results.values.keys():
                for field_id in self.profiling_results.values[gpu_id].keys():
                    field_name = DcgmFieldGetById(field_id).tag

                    field_avg_val = 0.0
                    num_vals = 0
                    for gpu_field_time in self.profiling_results.values[gpu_id][
                        field_id
                    ]:
                        if gpu_field_time.value is not None:
                            field_avg_val = (
                                field_avg_val * num_vals + gpu_field_time.value
                            ) / (num_vals + 1)
                            num_vals += 1
                    self.main_profiler.summary.append(
                        (f"GPU {gpu_id}, {field_name}({field_id})", f"{field_avg_val}")
                    )

            # Clear the profiling results to get ready for the next collection.
            self.profiling_results.EmptyValues()
